from ._utils import _developer_utils
import pandas as pd
import numpy as np
from scipy import stats
import seaborn as sns
import matplotlib.pyplot as plt

class AnomalyDetector:
    def __init__(self, f):
        self.f = f.__deepcopy__()

    def NaiveDetect(self, cilevel=0.99, **kwargs):
        """ Detects anomalies by breaking a series into its fundamental components:
        trend, seasonality, and residual. anomalies are defined as standard normal residuals
        further than a number of standard deviations away from the mean, determined by the value
        passed to cilevel. This is a simple, computationally cheap anomaly detector. Results
        are saved to the raw_anom and labeled_anom attributes.

        Args:
            cilevel (float): Default 0.99. The confidence interval used to determine how far
                away a given residual must be from the mean to be considered an anomaly. In a normal
                series that is decomposed effectively in this process, a cilevel of 0.95 would still expect
                to label 5% of its points as anomalies.
            **kwargs: Passed to the Forecaster.seasonal_decompose() method. 
                If extrapolate_trend is left unspecified, this will fail to produce results.
                See https://scalecast.readthedocs.io/en/latest/Forecaster/Forecaster.html#src.scalecast.Forecaster.Forecaster.seasonal_decompose.
        
        Returns:
            None

        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> detector = AnomalyDetector(f)
        >>> detector.NaiveDetect(extrapolate_trend='freq',train_only=True)
        """
        resid = self.f.seasonal_decompose(**kwargs).resid
        snr = (resid - resid.mean()) / resid.std()
        anom = snr.apply(lambda x: stats.norm.cdf(x))
        labeled_anom = anom.apply(
            lambda x: 1
            if x > cilevel + (1 - cilevel) / 2 or x < (1 - cilevel) / 2
            else 0
        )

        self.y = self.f.y.to_list()
        self.fvs = None
        self.lower = None
        self.upper = None
        self.raw_anom = anom
        self.labeled_anom = labeled_anom

    def EstimatorDetect(
        self,
        estimator,
        future_dates=None,
        cilevel=0.99,
        samples=100,
        return_fitted_vals=False,
        random_seed=None,
        **kwargs,
    ):
        """ Detects anomalies with one of a Forecaster object's estimators.
        An anomaly in this instance is defined as any value that falls
        out of the fitted values' bootstrapped confidence intervals
        determined by the value passed to cilevel. This can be a good method
        to detect anomalies if you want to attempt to break down a series'
        into trends, seasonalities, and autoregressive parts in a more complex
        manner than NaiveDetect would let you. It also gives access to RNN estimators,
        which are shown to be effective anomaly detectors for time series. Results
        are saved to the labeled_anom attribute.

        Args:
            estimator (str): One of `Forecaster.estimators`.
                The estimator to track anomalies with.
            future_dates (int): Optional. If this is specified with an integer, 
                the estimator will use that number of forecast steps. If you want 
                to span an entire series for anomalies, not just the training set, 
                future dates should be created either before initiating the AnomalyDetector 
                object or by passing an int to this arg. Future dates are what signal 
                to the object that we want to train the entire dataset.
            cilevel (float): Default 0.99. The confidence interval to use when
                bootstrapping confidence intervals.
            samples (int): Default 100. How many samples in the bootstrap to find confidence intervals.
            return_fitted_vals (bool): Default False. Whether to return a DataFrame
                of the fitted values and confidence intervals from the fitting process.
            random_seed (bool): Optional. Set a seed for consistent results.
            **kwargs: Passed to the Forecaster.manual_forecast() method.

        Returns:
            (DataFrame or None): A DataFrame of fitted values if return_fitted_vals is True.

        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        >>> f = transformer.DiffTransform(1)
        >>> f = transformer.DiffTransform(12)
        >>> detector = AnomalyDetector(f)
        >>> detector.EstimatorDetect(
        >>>    estimator='lstm',
        >>>    cilevel=.99,
        >>>    test_only=True,
        >>>    lags=24,
        >>>    epochs=25,
        >>>    validation_split=.2,
        >>>    shuffle=True,
        >>>    lstm_layer_sizes=(16,16,16),
        >>>    dropout=(0,0,0),
        >>> )
        """
        def _find_cis(f, samples):
            """ bootstraps the upper and lower forecast estimates using the info stored in cilevel and bootstrap_samples
            """
            y = f.y.values[-len(f.fitted_values):]
            fvs = f.fitted_values
            resids = [fv - ac for fv, ac in zip(fvs, y[-len(fvs) :])]
            bootstrapped_resids = np.random.choice(resids, size=samples)
            bootstrap_mean = np.mean(bootstrapped_resids)
            bootstrap_std = np.std(bootstrapped_resids)
            return _developer_utils._set_ci_step(
                f = f,
                s = bootstrap_std,
            ) + bootstrap_mean
        
        if random_seed is not None:
            random.seed(random_seed)
        
        f1 = self.f.__deepcopy__()
        call_me = estimator if "call_me" not in kwargs.keys() else kwargs["call_me"]

        if future_dates is None and not f1.future_dates.to_list():
            f1.generate_future_dates(1)  # because we have to have at least 1
        elif future_dates is not None:
            f1.generate_future_dates(future_dates)
        f1.set_estimator(estimator)
        f1.set_cilevel(cilevel)
        f1.manual_forecast(**kwargs)
        fvs = f1.export_fitted_vals(call_me).set_index("DATE")
        #fvs["range"] = f1.history[call_me]["CIPlusMinus"]
        fvs["range"] = _find_cis(
            f = f1,
            samples = samples,
        )
        fvs["labeled_anom"] = fvs[["Actuals", "FittedVals", "range"]].apply(
            lambda x: 1 if (x[1] > (x[0] + x[2])) | (x[1] < (x[0] - x[2])) else 0,
            axis=1,
        )
        self.y = fvs["Actuals"].to_list()
        self.fvs = fvs["FittedVals"]
        self.lower = fvs[["FittedVals", "range"]].apply(lambda x: x[0] - x[1], axis=1)
        self.upper = fvs[["FittedVals", "range"]].apply(lambda x: x[0] + x[1], axis=1)
        self.raw_anom = None
        self.labeled_anom = fvs["labeled_anom"]

        if return_fitted_vals:
            return fvs

    def MonteCarloDetect(self, start_at, stop_at, sims=100, cilevel=0.99):
        """ Detects anomalies by running a series of monte carlo simulations
        over a span of the series, using the observations before the span start 
        to determine the initial assumed distribution. Results are saved to the 
        raw_anom, labeled_anom, and mc_results attributes. It is a good idea to 
        transform the series before running so that it is stationary and not seasonal.
        In other words, the series distribution should be as close to normal as possible.
        
        Args:
            start_at (int, str, Datetime.Datetime, or pandas.Timestamp):
                If int, will start at that number obs in the series.
                Anything else should be a date-like object that can be
                parsed by the pandas.Timestamp() function, representing the
                starting point of the simulation. All observations before this
                point will be used to determine the mean/std of the intial distribution.
            stop_at (int, str, Datetime.Datetime, or pandas.Timestamp):
                If int, will stop at that number obs in the series.
                Anything else should be a date-like object that can be
                parsed by the pandas.Timestamp() function, representing the
                stopping point of the simulation.
            sims (int): The number of simulations.
            cilevel (float): Default .99.
                The percentage of points in the simulation that a given actual
                observation needs to be outside of the simulated series to be considered
                an anomaly. 

        Returns:
            None

        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        >>> f = transformer.DiffTransform(1)
        >>> f = transformer.DiffTransform(12)
        >>> detector = AnomalyDetector(f)
        >>> detector.MonteCarloDetect('2010-01-01','2020-12-01',cilevel=.99)
        """
        f1 = self.f.__deepcopy__()
        if not isinstance(start_at, int):
            start_at = f1.current_dates.to_list().index(pd.Timestamp(start_at))
        if not isinstance(stop_at, int):
            stop_at = f1.current_dates.to_list().index(pd.Timestamp(stop_at))
        stop_at = stop_at + 1
        results = pd.DataFrame(
            {
                "Date": f1.current_dates.to_list()[start_at:stop_at],
                "Actuals": f1.y.to_list()[start_at:stop_at],
            },
        ).set_index("Date")
        for s in range(sims):
            obs = [i for i in f1.y.to_list()[:start_at]]
            simmed_line = []
            for i in range(stop_at - start_at):
                simmed_line.append(
                    np.random.normal(loc=np.mean(obs), scale=np.std(obs))
                )
                obs.append(simmed_line[-1])
            df_tmp = pd.DataFrame({f"Iter{s}":simmed_line},index=results.index)
            results = pd.concat([results,df_tmp],axis=1)
            #results[f"Iter{s}"] = simmed_line
        results2 = results.copy()
        results2["sims_mean"] = results.drop("Actuals", axis=1).mean(axis=1)
        results2["sims_std"] = results.drop("Actuals", axis=1).std(axis=1)
        anom = results2[["Actuals", "sims_mean", "sims_std"]].apply(
            lambda x: stats.norm.cdf(x[0], loc=x[1], scale=x[2]), axis=1,
        )
        labeled_anom = anom.apply(
            lambda x: 1
            if x > cilevel + (1 - cilevel) / 2 or x < (1 - cilevel) / 2
            else 0
        )

        self.raw_anom = anom
        self.labeled_anom = labeled_anom
        self.y = results["Actuals"].to_list()
        self.fvs = None
        self.lower = None
        self.upper = None
        self.mc_results = results.reset_index()

    def MonteCarloDetect_sliding(
        self, historical_window, step, **kwargs,
    ):
        """ Detects anomalies by running a series of monte carlo simulations
        rolling over a span of the series. It is a good idea to 
        transform the series before running so that it is stationary and not seasonal.
        In other words, the series distribution should be as close to normal as possible.

        Args:
            historical_window (int): The number of periods to begin the initial search.
            step (int): How far to step forward after a scan.
            **kwargs: Passed to the `MonteCarloDetect()` method. `start_at` and `stop_at` passed
                automatically based on the values passed to the other arguments in this function.

        Returns:
            None

        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        >>> f = transformer.DiffTransform(1)
        >>> f = transformer.DiffTransform(12)
        >>> detector = AnomalyDetector(f)
        >>> detector.MonteCarloDetect_sliding(60,30)
        """
        raw_anom = pd.Series(dtype=float)
        labeled_anom = pd.Series(dtype=float)
        y = []
        n = len(self.f.y)
        y = self.f.y.to_list()[historical_window:]
        for end_idx in range(historical_window * 2, n + historical_window, step):
            print(f"scanning from obs {end_idx - historical_window} to obs {end_idx}")
            self.MonteCarloDetect(
                end_idx - historical_window, min(end_idx, n - 1), **kwargs
            )
            raw_anom = (
                pd.concat([raw_anom, self.raw_anom])
                .reset_index()
                .sort_values(["index", 0])
                .drop_duplicates(subset=["index"], keep="last")
                .set_index("index")[0]
            )
            labeled_anom = (
                pd.concat([labeled_anom, self.labeled_anom])
                .reset_index()
                .sort_values(["index", 0])
                .drop_duplicates(subset=["index"], keep="last")
                .set_index("index")[0]
            )

        self.raw_anom = raw_anom
        self.labeled_anom = labeled_anom
        self.y = y

    def WriteAnomtoXvars(self, f=None, future_dates=None, **kwargs):
        """ Writes the Xvars from the previously called anomaly detector to Xvars in
        a Forecaster object. Each anomaly is its own dummy variable on the date it is 
        found. A future distriution could detect level shifts.

        Args:
            f (Forecaster): optional. if you pass an object here,
                that object will receive the Xvars. otherwise,
                it will apply to the copy of the object stored in the
                the AnomalyDetector object when it was initialized.
                this Forecaster object is stored in the f attribute.
            future_dates (int): optional. if you pass a future dates
                length here, it will write that many dates to the
                Forecaster object and future anomaly variables will be
                passed as arrays of 0s so that any algorithm you train
                will be able to use them into future horizon.
            **kwargs: passed to the Forecaster.ingest_Xvars_df() function.
                see https://scalecast.readthedocs.io/en/latest/Forecaster/Forecaster.html#src.scalecast.Forecaster.Forecaster.ingest_Xvars_df
        Returns:
            (Forecaster) an object with the Xvars written.
        
        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        >>> f = transformer.DiffTransform(1)
        >>> f = transformer.DiffTransform(12)
        >>> detector = AnomalyDetector(f)
        >>> detector.MonteCarloDetect('2010-01-01','2020-12-01',cilevel=.99)
        >>> f = detector.WriteAnomtoXvars(drop_first=True)
        """
        f = self.f if f is None else f.deepcopy()
        if future_dates is not None:
            f.generate_future_dates(future_dates)
        df = pd.DataFrame(index=f.current_dates.to_list() + f.future_dates.to_list())
        df["Anomaly"] = self.labeled_anom
        df = df.reset_index()
        df.columns = ["Date", "Anomaly"]
        df["Anomaly"] = df[["Date", "Anomaly"]].apply(
            lambda x: x[0].strftime("%Y-%m-%d") if x[1] == 1 else "No", axis=1
        )
        df["Anomaly"] = pd.Categorical(
            df["Anomaly"].to_list(),
            categories=["No"] + [c for c in df["Anomaly"].unique() if c != "No"],
            ordered=True,
        )  # all this to make "No" the first cat
        f.ingest_Xvars_df(df, date_col="Date", **kwargs)
        return f

    def adjust_anom(self, f=None, method="q", q=10):
        """ Changes the values of identified anomalies and returns a Forecaster object.

        Args: 
            f (Forecaster): Optional. If you pass an object here,
                that object will have its y values altered. Otherwise,
                it will apply to the copy of the object stored in the
                the AnomalyDetector object when it was initialized.
                this Forecaster object is stored in the f attribute.
            method (str): The following methods are supported: "q" and "interpolate".
                "q" uses q-cutting from pandas and fills values with second-to-last
                q value in either direction. For example, if q == 10, then high anomaly
                values will be replaced with the 90th percentile of the rest of the series
                data. Low anomaly values will be replaced with the 10th percentile of the 
                rest of the series. This is a good method for when your data is stationary.
                For data with a trend, 'interpolate' is better as it fills in values linearly
                based on the values before and after consecutive anomaly values. Be careful
                when using "q" with differenced data. when undifferencing,
                original values will be reverted back to.
            q (int): Default 10. 
                The q-value to use when method == 'q'.
                Ignored when method != 'q'.

        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> detector = AnomalyDetector(f)
        >>> detector.EstimatorDetect(
        >>>    estimator='arima',
        >>>    order=(1,1,1),
        >>>    seasonal_order=(1,1,1),
        >>> )
        >>> f = detector.adjust_anom(method='interpolate')
        """
        f = self.f if f is None else f.deepcopy()
        df = pd.DataFrame({"y": f.y.to_list()}, index=f.current_dates.to_list(),)
        df["a"] = self.labeled_anom

        if method == "q":
            df["r"] = pd.qcut(df["y"], q=q, labels=list(range(q)))

            top = df.loc[df["r"] == (q - 1), "y"].min()
            bottom = df.loc[df["r"] == 0, "y"].max()

            for i, v in df.reset_index().iterrows():
                if v["a"] == 1:
                    f.y.values[i] = top if v["y"] > df["y"].mean() else bottom
        elif method == "interpolate":
            df["anom_num"] = 0
            df["a_1"] = df["a"].shift()
            df["a_-1"] = df["a"].shift(-1)
            df = df.fillna(0)
            df.loc[(df["a_-1"] == 1) & (df["a"] == 0), "anom_num"] = 1
            df["anom_num"] = df["anom_num"].cumsum()
            df.loc[
                ~(
                    ((df["a_-1"] == 1) & (df["a"] == 0))
                    | (df["a"] == 1)
                    | ((df["a_1"] == 1) & (df["a"] == 0))
                ),
                "anom_num",
            ] = 0

            for anom in df["anom_num"].unique():
                if anom == 0:
                    continue

                df_tmp = df.loc[df["anom_num"] == anom]
                idx = df_tmp.iloc[1:-1, :].index.to_list()
                slope = (df_tmp.iloc[-1, 0] - df_tmp.iloc[0, 0]) / (df_tmp.shape[0] - 1)

                for i, d in enumerate(idx):
                    df.loc[d, "y"] = df_tmp.iloc[0, 0] + slope * (i + 1)

            f.y = df["y"].reset_index(drop=True)
        else:
            raise ValueError(f'method arg expected "q" or "interpolate", got {method}')

        return f

    def plot_anom(self, label=True, strftime_fmt="%Y-%m-%d"):
        """ Plots the series used to detect anomalies and red dashes around points that
        were identified as anomalies from the last algorithm run.

        Args:
            label (bool): Default True.
                Whether to add the date label to each plotted point.
            strftime_fmt (str): Default '%Y-%m-%d'.
                The string format to convert dates to when label is True.
                When label is False, this is ignored.

        Returns:
            (Axis): The figure's axis.

        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> import matplotlib.pyplot as plt
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        >>> f = transformer.DiffTransform(1)
        >>> f = transformer.DiffTransform(12)
        >>> detector = AnomalyDetector(f)
        >>> detector.MonteCarloDetect('2010-01-01','2020-12-01',cilevel=.99)
        >>> ax = f.plot_anom()
        >>> plt.show()
        """
        ax = self.f.plot(models=None)

        for y, a, d in zip(self.y, self.labeled_anom, self.labeled_anom.index):
            if a == 1:
                ax.scatter(d, y, marker="_", color="red", linewidths=2, label=None)
                if label:
                    ax.text(
                        d,
                        y * 1.05,
                        d.strftime(strftime_fmt),
                        color="red",
                        size=11,
                        label=None,
                    )
        if self.fvs is not None:
            ax.plot(
                self.labeled_anom.index,
                self.fvs,
                color="#FFA500",
                alpha=0.5,
                label="Fitted Vals",
            )
            ax.fill_between(
                x=self.labeled_anom.index,
                y1=self.upper,
                y2=self.lower,
                alpha=0.2,
                color="#FFA500",
                label="Conf. Interval",
            )

        ax.legend()
        return ax

    def plot_mc_results(self,ax=None,figsize=(12,6)):
        """ Plots the results from a monte-carlo detector: the series' original values
        and the simulated lines.

        Args:
            ax (Axis): Optional. An existing axis to display the figure on.
            figsize (tuple): Default (12,6). The size of the resulting figure.
                Ignored if axis is not None.

        Returns:
            (Axis): The figure's axis.

        >>> from scalecast.AnomalyDetector import AnomalyDetector
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> from scalecast.Forecaster import Forecaster
        >>> import pandas_datareader as pdr
        >>> import matplotlib.pyplot as plt
        >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
        >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        >>> f = transformer.DiffTransform(1)
        >>> f = transformer.DiffTransform(12)
        >>> detector = AnomalyDetector(f)
        >>> detector.MonteCarloDetect('2010-01-01','2020-12-01',cilevel=.99)
        >>> ax = f.plot_mc_results()
        >>> plt.show()
        """
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        results = self.mc_results
        sns.lineplot(x="Date", y="Actuals", data=results, label="actuals", ax=ax)
        for c in results.iloc[:, 2:]:
            ax.plot(results["Date"], results[c], alpha=0.2)
        return ax



from darts.utils.utils import SeasonalityMode, TrendMode, ModelMode

theta = {
    'theta':[0.5,1,1.5,2,2.5,3],
    'model_mode':[
        ModelMode.ADDITIVE,
        ModelMode.MULTIPLICATIVE
    ],
    'season_mode':[
        SeasonalityMode.MULTIPLICATIVE,
        SeasonalityMode.ADDITIVE
    ],
    'trend_mode':[
        TrendMode.EXPONENTIAL,
        TrendMode.LINEAR
    ],
}


from ._Forecaster_parent import _tune_test_forecast
from .Forecaster import Forecaster
from .MVForecaster import MVForecaster
import typing
from typing import Dict, Union
from ipywidgets import widgets
from IPython.display import display, clear_output
import matplotlib.pyplot as plt
import seaborn as sns

def results_vis(
    f_dict: Dict[str,Forecaster],
    plot_type: str = "forecast",
    include_train: Union[bool, int] = True,
    figsize = (12,6),
):
    """ Visualize the forecast results from many different Forecaster objects leveraging Jupyter widgets.

    Args:
        f_dict (dict[str,Forecaster]): Dictionary of forcaster objects.
            Works best if two or more models have been evaluated in each dictionary value.
        plot_type (str): One of {"forecast","test"}. Default "forecast".
            The type of results to visualize.
        include_train (bool or int): Optional.
            Whether to include the complete training set in the plot or how many traning-set observations to include.
            Passed to include_train parameter when plot_type = 'test'.
            Ignored when plot_type = 'forecast'.
        figsize (tuple): Default (12,6). Size of the resulting figure.

    Returns:
        None 
    """
    if plot_type not in {"forecast", "test"}:
        raise ValueError(f'plot_type must be "forecast" or "test", got {plot_type}')

    def display_user_selections(
        ts_selection, mo_selection, lv_selection, ci_selection, me_selection
    ):
        selected_data = f_dict[ts_selection]
        if plot_type == "forecast":
            selected_data.plot(
                models=f"top_{mo_selection}",
                order_by=me_selection,
                level=lv_selection,
                ci=ci_selection,
                figsize=figsize,
            )
        else:
            selected_data.plot_test_set(
                models=f"top_{mo_selection}",
                order_by=me_selection,
                include_train=include_train,
                level=lv_selection,
                ci=ci_selection,
                figsize=figsize,
            )
        plt.title(ts_selection + " Forecast Results", size=16)
        plt.show()

    def on_button_clicked(b):
        mo_selection = mo_dd.value
        ts_selection = ts_dd.value
        lv_selection = lv_dd.value
        ci_selection = ci_dd.value
        me_selection = me_dd.value
        with output:
            clear_output()
            display_user_selections(
                ts_selection, mo_selection, lv_selection, ci_selection, me_selection
            )

    all_models = []
    for k, f in f_dict.items():
        all_models += [fcst for fcst in f.history.keys() if fcst not in all_models]
    ts_dd = widgets.Dropdown(options=f_dict.keys(), description="Time Series:")
    mo_dd = widgets.Dropdown(
        options=range(1, len(all_models) + 1), description="No. Models"
    )
    lv_dd = widgets.Dropdown(options=[False, True], description="View Level")
    ci_dd = widgets.Dropdown(
        options=[False, True], description="View Confidence Intervals"
    )
    me_dd = widgets.Dropdown(
        options=sorted(f.determine_best_by), # f will be last object iterated through above
        description="Order By", 
        value = [m for m in f.determine_best_by if m.startswith('TestSet')][0],
    )

    # never changes
    button = widgets.Button(description="Select Time Series")
    output = widgets.Output()

    display(ts_dd, mo_dd, lv_dd, ci_dd, me_dd)
    display(button, output)

    button.on_click(on_button_clicked)


def results_vis_mv(
    f_dict: Dict[str,MVForecaster], 
    plot_type="forecast", 
    include_train=True, 
    figsize = (12,6)
):
    """ Visualize the forecast results from many different MVForecaster objects leveraging Jupyter widgets.

    Args:
        f_dict (dict[str,MVForecaster]): Dictionary of forcaster objects.
            Works best if two or more models have been evaluated in each dictionary value.
        plot_type (str): One of {"forecast","test"}. Default "forecast".
            The type of results to visualize.
        include_train (bool or int): Optional.
            Whether to include the complete training set in the plot or how many traning-set observations to include.
            Passed to include_train parameter when plot_type = 'test'.
            Ignored when plot_type = 'forecast'.
        figsize (tuple): Default (12,6). Size of the resulting figure.

    Returns:
        None
    """
    if plot_type not in {"forecast", "test"}:
        raise ValueError(f'plot_type must be "forecast" or "test", got {plot_type}')

    def display_user_selections(
        mo_selection, ts_selection, lv_selection, ci_selection, se_selection
    ):
        selected_data = f_dict[ts_selection]
        if plot_type == "forecast":
            selected_data.plot(
                models=mo_selection,
                series=se_selection,
                level=lv_selection,
                ci=ci_selection,
                figsize=figsize,
            )
        else:
            selected_data.plot_test_set(
                models=mo_selection,
                series=se_selection,
                level=lv_selection,
                ci=ci_selection,
                include_train=include_train,
                figsize=figsize,
            )
        plt.title(ts_selection + " Forecast Results", size=16)
        plt.show()

    def on_button_clicked(b):
        mo_selection = mo_se.value
        ts_selection = ts_dd.value
        lv_selection = lv_dd.value
        ci_selection = ci_dd.value
        se_selection = se_se.value
        with output:
            clear_output()
            display_user_selections(
                mo_selection, ts_selection, lv_selection, ci_selection, se_selection
            )

    all_models = []
    n_series = 2
    for k, f in f_dict.items():
        all_models += [fcst for fcst in f.history.keys() if fcst not in all_models]
        n_series = max(n_series, f.n_series)
    series = [f"series{i+1}" for i in range(n_series)]
    ts_dd = widgets.Dropdown(options=f_dict.keys(), description="Time Series:")
    mo_se = widgets.SelectMultiple(
        options=all_models, description="Models", selected=all_models
    )
    se_se = widgets.SelectMultiple(
        options=series, description="Series", selected=series
    )
    lv_dd = widgets.Dropdown(options=[False, True], description="View Level")
    ci_dd = widgets.Dropdown(
        options=[False, True], description="View Confidence Intervals"
    )

    # never changes
    button = widgets.Button(description="Select Time Series")
    output = widgets.Output()

    display(ts_dd, se_se, mo_se, lv_dd, ci_dd)
    display(button, output)

    button.on_click(on_button_clicked)


def tune_test_forecast(
    f,
    models,
    cross_validate=False,
    dynamic_tuning=False,
    dynamic_testing=True,
    summary_stats=False,
    feature_importance=False,
    fi_method="pfi",
    limit_grid_size=None,
    suffix=None,
    error='raise',
    **cvkwargs,
):
    """ Tunes, tests, and forecasts a series of models with a progress bar through tqdm.

    Args:
        f (Forecaster or MVForecaster): The object to run the models through.
        models (list-like):
            Each element must be in Forecaster.can_be_tuned.
        cross_validate (bool): Default False.
            Whether to tune the model with cross validation. 
            If False, uses the validation slice of data to tune.
        dynamic_tuning (bool or int): Default False.
            Whether to dynamically tune the model or, if int, how many forecast steps to dynamically tune it.
        dynamic_testing (bool or int): Default True.
            Whether to dynamically/recursively test the forecast (meaning AR terms will be propogated with predicted values).
            If True, evaluates recursively over the entire out-of-sample slice of data.
            If int, window evaluates over that many steps (2 for 2-step recurvie testing, 12 for 12-step, etc.).
            Setting this to False or 1 means faster performance, 
            but gives a less-good indication of how well the forecast will perform more than one period out.
        summary_stats (bool): Default False.
            Whether to save summary stats for the models that offer those.
            Does not work for `MVForecaster` objects.
        feature_importance (bool): Default False.
            Whether to save permutation feature importance information for the models that offer those.
            Does not work for `MVForecaster` objects.
        fi_method (str): One of {'pfi','shap'}. Default 'pfi'.
            The type of feature importance to save for the models that support it.
            Ignored if feature_importance is False.
            Does not work for `MVForecaster` objects.
        limit_grid_size (int or float): Optional. Pass an argument here to limit each of the grids being read.
            See https://scalecast.readthedocs.io/en/latest/Forecaster/Forecaster.html#src.scalecast.Forecaster.Forecaster.limit_grid_size.
        suffix (str): Optional. A suffix to add to each model as it is evaluate to differentiate them when called
            later. If unspecified, each model can be called by its estimator name.
        error (str): One of 'ignore','raise','warn'; default 'raise'.
            What to do with the error if a given model fails.
            'warn' prints a warning that the model could not be evaluated.
        **cvkwargs: Passed to the cross_validate() method.

    Returns:
        None
    """
    _tune_test_forecast(
        f=f,
        models=models,
        cross_validate=cross_validate,
        dynamic_tuning=dynamic_tuning,
        dynamic_testing=dynamic_testing,
        limit_grid_size=limit_grid_size,
        suffix=suffix,
        error=error,
        summary_stats=summary_stats,
        feature_importance=feature_importance,
        fi_method=fi_method,
        tqdm = True,
    )



from .Forecaster import Forecaster
from .SeriesTransformer import SeriesTransformer
import numpy as np
import pandas as pd
from typing import List, Tuple, Union, Dict
import typing

class Transformer:
    def __init__(self,transformers: List[Tuple]):
        """ Initiates the transformer pipeline.

        Args:
            transformers (list[tuple]): A list of transformations to apply to the time series stored in a Forecaster object.
                The tuple's first element should match the name of a transform function from the SeriesTransformer object: 
                https://scalecast.readthedocs.io/en/latest/Forecaster/SeriesTransformer.html.
                Positional and keyword arguments can be passed to these functions.
                If a given tuple is more than 1 in length, the `fit_transform()` method will parse
                Elements after index 0 as positional arguments.
                Keywords are passed as a dictionary in the last position of tuples greater than 1 in length. 
                Therefore, if the last argument in the tuple is a dict type,
                This is assumed to contain the keyword arguments. 
                If the last positional argument you wish to pass happens to be dict type,
                you can eaither pass it as a keyword argument or place an additional (empty) dictionary at the end of the tuple.

        >>> from scalecast.Pipeline import Transformer
        >>> transformer = Transformer(
        >>>     transformers = [
        >>>         ('LogTransform',),
        >>>         ('DiffTransform',1),
        >>>         ('DiffTransform',12),
        >>>     ],
        >>> )
        """
        # validate types (transform str types to tuple)
        for i, transformer in enumerate(transformers):
            if isinstance(transformer,str):
                transformers[i] = (transformer,)
            elif not isinstance(transformer,tuple):
                raise TypeError(f'Expected elements of transformer list to be tuple type, got {type(transformer)}.')

        self.transformers = transformers

    def __repr__(self):
        return (
            "Transformer(\n"
            "  transformers = [\n"
            "    {}\n"
            "  ]\n"
            ")".format(",\n    ".join([str(i) for i in self.transformers]))
        )

    def __str__(self):
        return self.__repr__()

    def fit_transform(self,f: Forecaster) -> Forecaster:
        """ Applies the transformation to the series stored in the Forecaster object.

        Args:
            f (Forecaster): The Forecaster object that stores the series that will be transformed.

        Returns:
            (Forecaster): A Forecaster object with the transformed series.

        >>> from scalecast.Pipeline import Transformer
        >>> transformer = Transformer(
        >>>     transformers = [
        >>>         ('LogTransform',),
        >>>         ('DiffTransform',1),
        >>>         ('DiffTransform',12),
        >>>     ],
        >>> )
        >>> f = transformer.fit_transform(f)
        """
        for i, transformer in enumerate(self.transformers):
            if len(transformer) > 1:
                args = [i for i in transformer[1:-1]]
                args += transformer[-1:] if not isinstance(transformer[-1],dict) else []
                kwargs = transformer[-1] if isinstance(transformer[-1],dict) else {}
            else:
                args = []
                kwargs = {}
            
            if i == 0:
                self.base_transformer = SeriesTransformer(f)
            
            f = getattr(self.base_transformer,transformer[0])(*args,**kwargs)
        return f

class Reverter:
    def __init__(self,
        reverters: List[Tuple],
        base_transformer: Union[Transformer,SeriesTransformer]
    ):
        """ Initiates the reverter pipeline.

        Args:
            reverters (list[tuple]): A list of revert funcs to apply to the time series stored in a Forecaster object.
                The tuple's first element should match the name of a revert function from the SeriesTransformer object: 
                https://scalecast.readthedocs.io/en/latest/Forecaster/SeriesTransformer.html.
                Positional and keyword arguments can be passed to these functions.
                If a given tuple is more than 1 in length, the `fit_transform()` method will parse
                elements after index 0 as positional arguments.
                Keywords are passed as a dictionary in the last position of tuples greater than 1 in length. 
                Therefore, if the last argument in the tuple is a dict type,
                this is assumed to contain the keyword arguments. 
                If the last positional argument you wish to pass happens to be dict type,
                You can eaither pass it as a keyword argument or place an additional (empty) dictionary at the end of the tuple.
            base_transformer (Transformer|SeriesTransformer): The object that was used to make the original transformations.
                These objects contain the key information to undifference and unscale the stored data 
                and therefore this argument is required.

        >>> from scalecast.Pipeline import Reverter
        >>> reverter = Reverter(
        >>>     reverters = [
        >>>         ('DiffRevert',12),
        >>>         ('DiffRevert',1),
        >>>         ('LogRevert',),
        >>>     ],
        >>>     base_transformer = transformer,
        >>> )
        """
        # validate types (transform str types to tuple)
        for i, reverter in enumerate(reverters):
            if isinstance(reverter,str):
                reverters[i] = (reverter,)
            elif not isinstance(reverter,tuple):
                raise TypeError(f'Expected elements of reverter list to be tuple type, got {type(reverter)}.')
        
        self.reverters = reverters
        self.base_transformer = base_transformer

    def __repr__(self):
        return (
            "Reverter(\n"
            "  reverters = [\n"
            "    {}\n"
            "  ],\n"
            "  base_transformer = {}\n"
            ")".format(
                ",\n    ".join([str(i) for i in self.reverters]),
                self.base_transformer,
            )
        )

    def __str__(self):
        return self.__repr__()

    def fit_transform(self,f: Forecaster) -> Forecaster:
        """ Applies the revert function to the series stored in the Forecaster object.

        Args:
            f (Forecaster): The Forecaster object that stores the series that will be reverted.

        Returns:
            (Forecaster): A Forecaster object with the reverted series.

        >>> from scalecast.Pipeline import Reverter
        >>> reverter = Reverter(
        >>>     reverters = [
        >>>         ('DiffRevert',12),
        >>>         ('DiffRevert',1),
        >>>         ('LogRevert',),
        >>>     ],
        >>>     base_transformer = transformer,
        >>> )
        >>> f = reverter.fit_transform(f)
        """
        base_transformer = (
            self.base_transformer if not hasattr(self.base_transformer,'base_transformer')
            else self.base_transformer.base_transformer
        )
        for reverter in self.reverters:
            if len(reverter) > 1:
                args = [i for i in reverter[1:-1]]
                args += reverter[-1:] if not isinstance(reverter[-1],dict) else []
                kwargs = reverter[-1] if isinstance(reverter[-1],dict) else {}
            else:
                args = []
                kwargs = {}

            f = getattr(base_transformer,reverter[0])(*args,**kwargs)
        return f

class Pipeline_parent:
    def _prepare_backtest(
        self,
        *fs,
        n_iter,
        jump_back,
        series_length,
        fcst_length,
        test_length,
        cis,
        cilevel,
    ) -> List[List[Tuple[Forecaster,np.array]]]:
        results = []
        for h, f in enumerate(fs):
            fcst_length = len(f.future_dates) if fcst_length is None else fcst_length
            results.append([])
            for i in range(n_iter):
                hold_out_len = ((i+1) * (fcst_length)) + (i * jump_back)
                hold_out = f.y.values[-hold_out_len :][: fcst_length]
                f1 = Forecaster(
                    y = f.y.values[: -hold_out_len],
                    current_dates = f.current_dates[: -hold_out_len],
                    future_dates = len(f.future_dates) if fcst_length is None else fcst_length,
                    test_length = f.test_length if test_length is None else test_length,
                    cis = f.cis if cis is None else cis,
                    cilevel = f.cilevel if cilevel is None else cilevel,
                )
                if series_length is not None:
                    f1.keep_smaller_history(series_length)
                results[h].append((f1,hold_out))
        return results

    def backtest(
        self,        
        *fs,
        n_iter=5,
        jump_back=1,
        series_length=None,
        fcst_length=None,
        test_length=None,
        cis=None,
        cilevel=None,
        **kwargs,
    ) -> List[Dict[str,pd.DataFrame]]:
        """ Runs an out-of-sample backtest of the pipeline over a certain amount of iterations.

        Args:
            *fs (Forecaster): Send one if univariate forecasting with the `Pipeline` class, 
                more than one if multivariate forecasting with the `MVPipeline` class.
            n_iter (int): Default 5. How many backtest iterations to perform.
            jump_back (int): Default 1. The space between consecutive training sets.
            series_length (int): Optional. The total length of each traning set. 
                Leave unspecified if you want to use every available training observation for each iteration.
            fcst_length (int): Optional. The forecast horizon length to forecast over for each iteration.
                Leave unspecified if you want to use the forecast horizon already programmed into the `Forecaster` object.
            test_length (int): Optional. The test set to hold out for each model evaluation.
                Leave unspecified if you want to use the test length already programmed into the `Forecaster` object.
            cis (bool): Optional. Whether to backtest confidence intervals. 
                Leave unspecified if you want to use whatever is already programmed into the `Forecaster` object.
            cilevel (float): Optional. What level to evaluate confidence intervals at.
                Leave unspecified if you want to use whatever is already programmed into the `Forecaster` object.
            **kwargs: Passed to the `fit_predict()` method from `Pipeline` or `MVPipeline`.

        Returns:
            (List[Dict[str,pd.DataFrame]]): The results from each model and backtest iteration.
            Each dict element of the resulting list corresponds to the Forecaster objects in the order
            they were passed (will be length 1 if univariate forecasting). Each key of each dict is either 'Actuals', 'Obs',
            or the name of a model that got backtested. Each value is a DataFrame with the iteration values.
            The 'Actuals' frame has the date information and are the actuals over each forecast horizon. 
            The 'Obs' frame has the actual historical observations to make each forecast, back padded with NA values to make each array the same length.

        >>> # univariate forecasting
        >>> pipeline = Pipeline(
        >>>     steps = [
        >>>         ('Transform',transformer),
        >>>         ('Forecast',forecaster),
        >>>         ('Revert',reverter),
        >>>     ],
        >>> )
        >>> backtest_results = pipeline.backtest(f,models=models)
        >>>
        >>> # multivariate forecasting
        >>> pipeline = MVPipeline(
        >>>    steps = [
        >>>        ('Transform',[transformer1,transformer2,transformer3]),
        >>>        ('Select Xvars',[auto_Xvar_select]*3),
        >>>        ('Forecast',forecaster,),
        >>>        ('Revert',[reverter1,reverter2,reverter3]),
        >>>    ],
        >>>    names = ['UTUR','UTPHCI','UNRATE'], # used to combine to the mvf object
        >>>    merge_Xvars = 'i', # used to combine to the mvf object
        >>> )
        >>> backtest_results = pipeline.backtest(f1,f2,f3)
        """
        results = []
        _prepare_backtest_results = self._prepare_backtest(
            *fs,
            n_iter=n_iter,
            jump_back=jump_back,
            series_length=series_length,
            fcst_length=fcst_length,
            test_length=test_length,
            cis=cis,
            cilevel=cilevel,
        )
        for res in _prepare_backtest_results:
            results.append({'Actuals':pd.DataFrame()})
            results[-1]['Obs'] = pd.DataFrame()
            for i, f in enumerate(res):
                results[-1]['Actuals'][f'Iter{i}Dates'] = f[0].future_dates.values
                results[-1]['Actuals'][f'Iter{i}Vals'] = f[1]
                if i == 0:
                    results[-1]['Obs'][f'Iter{i}'] = f[0].y.to_list()
                else:
                    results[-1]['Obs'][f'Iter{i}'] = (
                        [np.nan] * 
                        (
                            results[-1]['Obs'].shape[0] - len(f[0].y)
                        ) + f[0].y.to_list() 
                    )
        for i, _ in enumerate(_prepare_backtest_results[0]):
            fs = self.fit_predict(*[ft[i][0] for ft in _prepare_backtest_results],**kwargs)
            if not isinstance(fs,tuple):
                fs = (fs,)
            for k, f in enumerate(fs):
                for m,v in f.history.items():
                    if i == 0:
                        results[k][m] = pd.DataFrame({'Iter0Fcst':v['Forecast']})
                    else:
                        results[k][m][f'Iter{i}Fcst'] = v['Forecast']
                    if f.cis:
                        results[k][m][f'Iter{i}Lower'] = v['LowerCI']
                        results[k][m][f'Iter{i}Upper'] = v['UpperCI']

        return results

class Pipeline(Pipeline_parent):
    def __init__(self,steps: List[Tuple[str,Union[Transformer,Reverter,'function']]]):
        """ Initiates the full pipeline.

        Args:
            steps (list[tuple]): A list of transform, forecast, and revert funcs to apply
                to a Forecaster object. The first element of each tuple names the step.
                The second element should either be a Transformer or Reverter type or a function.
                If it is a function, the first argument in the function should require a Forecaster object.
                Functions are checked for as objects that do not have the `fit_transform()` method,
                so adding more elements to the Pipeline may be possible if they have a `fit_transform()` method.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.Pipeline import Transformer, Reverter, Pipeline
        >>> import pandas_datareader as pdr
        >>> 
        >>> models = ['mlr','elasticnet']
        >>> def forecaster(f,models):
        >>>     f.add_covid19_regressor()
        >>>     f.auto_Xvar_select(cross_validate=True)
        >>>     f.tune_test_forecast(models)
        >>>
        >>> df = pdr.get_data_fred(
        >>>     'HOUSTNSA',
        >>>     start='1959-01-01',
        >>>     end='2022-08-01'
        >>> )
        >>> f = Forecaster(
        >>>     y=df['HOUSTNSA'],
        >>>     current_dates=df.index,
        >>>     future_dates=24,
        >>> )
        >>> f.set_test_length(0.2)
        >>> f.set_validation_length(24)
        >>> transformer = Transformer(
        >>>     transformers = [
        >>>         ('LogTransform',),
        >>>         ('DiffTransform',1),
        >>>         ('DiffTransform',12),
        >>>     ],
        >>> )
        >>> reverter = Reverter(
        >>>     reverters = [
        >>>         ('DiffRevert',12),
        >>>         ('DiffRevert',1),
        >>>         ('LogRevert',),
        >>>     ],
        >>>     base_transformer = transformer,
        >>> )
        >>> pipeline = Pipeline(
        >>>     steps = [
        >>>         ('Transform',transformer),
        >>>         ('Forecast',forecaster),
        >>>         ('Revert',reverter),
        >>>     ],
        >>> )
        """
        # validate we have tuples
        for step in steps:
            if not isinstance(step,tuple):
                raise TypeError(f'Expected elements of pipeline steps list to be tuple type, got {type(step)}.')
        
        self.steps = steps

    def __repr__(self):
        return (
            "Pipeline(\n"
            "  steps = [\n"
            "    {}\n"
            "  ]\n"
            ")".format(",\n    ".join([str(i) for i in self.steps]))
        )

    def __str__(self):
        return self.__repr__()

    def fit_predict(self,f: Forecaster,**kwargs) -> Forecaster:
        """ Applies the transform, forecast, and revert functions to the series stored in the Forecaster object.

        Args:
            f (Forecaster): The Forecaster object that stores the series that will be sent through the pipeline.
            **kwargs: Passed to any 'function' types passed in the pipeline.

        Returns:
            (Forecaster): A Forecaster object with the stored results from the pipeline run.
        
        >>> pipeline = Pipeline(
        >>>     steps = [
        >>>         ('Transform',transformer),
        >>>         ('Forecast',forecaster),
        >>>         ('Revert',reverter),
        >>>     ],
        >>> )
        >>> f = pipeline.fit_predict(f,models=models)
        """
        for step in self.steps:
            func = step[1]
            if hasattr(func,'fit_transform'):
                f = func.fit_transform(f)
            else:
                func(f,**kwargs)
        return f

class MVPipeline(Pipeline_parent):
    def __init__(
        self,
        steps: List[Tuple[str,Union[List[Transformer],List[Reverter],'function']]],
        **kwargs,
    ):
        """ Initiates the full pipeline for multivariate forecasting applications.

        Args:
            steps: (list[tuple]): A list of transform, forecast, and revert funcs to apply
                to multiple Forecaster objects. The first element of each tuple names the step.
                The second element should be a list of Transformer objects, a list of Reverter objects,
                a list of functions, or a single function. If it is a function or list of functions, 
                the first argument in the should require a Forecaster or MVForecaster object.
                If it is a list of functions, Transformer, or Revereter objects,
                each one of these will be called on the Forecaster objects in the order they are passed
                to the `fit_predict()` method.
                Functions are checked for as objects that do not have the `fit_transform()` method,
                so adding more elements to the Pipeline may be possible if they have a `fit_transform()` method.
            **kwargs: Passed to MVForecaster(). See
                https://scalecast.readthedocs.io/en/latest/Forecaster/MVForecaster.html#src.scalecast.MVForecaster.MVForecaster.__init__.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.Pipeline import MVPipeline
        >>> from scalecast.util import pdr_load, find_optimal_transformation
        >>> 
        >>> def auto_Xvar_select(f):
        >>>    f.auto_Xvar_select(max_ar=0)
        >>> def forecaster(mvf):
        >>>     mvf.set_test_length(24)
        >>>     mvf.set_estimator('elasticnet')
        >>>     mvf.manual_forecast(alpha=.2,lags=12)
        >>>
        >>> f1 = pdr_load('UTUR',future_dates=24,start='1970-01-01',end='2022-07-01')
        >>> f2 = pdr_load('UTPHCI',future_dates=24,start='1970-01-01',end='2022-07-01')
        >>> f3 = pdr_load('UNRATE',future_dates=24,start='1970-01-01',end='2022-07-01')
        >>> # doing this helps the `DetrendTransform()` function
        >>> fs = [f1,f2,f3]
        >>> for f in fs:
        >>>     f.set_test_length(24)
        >>>
        >>> transformer1, reverter1 = find_optimal_transformation(f1)
        >>> transformer2, reverter2 = find_optimal_transformation(f2)
        >>> transformer3, reverter3 = find_optimal_transformation(f3)
        >>> 
        >>> pipeline = MVPipeline(
        >>>     steps = [
        >>>         ('Transform',[transformer1,transformer2,transformer3]),
        >>>         ('Select Xvars',[auto_Xvar_select]*3), # finds xvars for each object
        >>>         ('Forecast',forecaster,), # combines to an mvf object
        >>>         ('Revert',[reverter1,reverter2,reverter3]), # breaks back to f objects
        >>>     ],
        >>>     names = ['UTUR','UTPHCI','UNRATE'],
        >>>     merge_Xvars = 'i',
        >>> )
        """
        for step in steps:
            if not isinstance(step,tuple):
                raise TypeError(f'Expected elements of pipeline steps list to be tuple type, got {type(step)}.')
        
        self.steps = steps
        self.kwargs = kwargs

    def __repr__(self):
        return (
            "MVPipeline(\n"
            "  steps = [\n"
            "    {}\n"
            "  ]\n"
            ")".format(",\n    ".join([str(i) for i in self.steps]))
        )

    def __str__(self):
        return self.__repr__()

    def fit_predict(self,*fs: Forecaster,**kwargs):
        """ Applies the transform, forecast, and revert functions to the series stored in the Forecaster object.
        The order of Forecaster passed to *fs is the order all functions in lists will be applied.

        Args:
            *fs (Forecaster): The Forecaster objects that stores the series that will be sent through the pipeline.
            **kwargs: Passed to any 'function' types passed in the pipeline.

        Returns:
            (Tuple[Forecaster] | MVForecaster): If the last element in the pipeline is a list of reverter functions
            this function returns the individual Forecaster objects. If not, an MVForecaster object is returned.
        
        >>> pipeline = MVPipeline(
        >>>    steps = [
        >>>        ('Transform',[transformer1,transformer2,transformer3]),
        >>>        ('Select Xvars',[auto_Xvar_select]*3), # applied to Forecaster objects
        >>>        ('Forecast',forecaster,), # combines to an mvf object and calls the function
        >>>        ('Revert',[reverter1,reverter2,reverter3]), # breaks back to f objects
        >>>    ],
        >>>    names = ['UTUR','UTPHCI','UNRATE'], # used to combine to the mvf object
        >>>    merge_Xvars = 'i', # used to combine to the mvf object
        >>> )
        >>> f1, f2, f3 = pipeline.fit_predict(f1,f2,f3)
        """
        from .MVForecaster import MVForecaster
        from .util import break_mv_forecaster
        from .multiseries import keep_smallest_first_date

        if 'not_same_len_action' not in kwargs:
            keep_smallest_first_date(*fs)
        elif kwargs['not_same_len_action'] == 'trim':
            keep_smallest_first_date(*fs)

        i = 0
        fs = list(fs)
        for step in self.steps:
            func_list = step[1]
            if hasattr(func_list,'__len__'):
                if len(fs) != len(func_list):
                    raise ValueError('Must pass as many functions as there are Forecaster objects.')
                if hasattr(func_list[0],'fit_transform'):
                    if i == 1:
                        fs = list(break_mv_forecaster(mvf))
                        i += 1
                    for idx, func in enumerate(zip(fs,func_list)):
                        if func[1] is not None:
                            if i == 2: # reverting
                                if hasattr(func[1].base_transformer,'base_transformer'):
                                    func[1].base_transformer.base_transformer.f = func[0]
                                else:
                                    func[1].base_transformer.f = func[0]
                            fs[idx] = func[1].fit_transform(func[0])
                else:
                    for f, func in zip(fs,func_list):
                        if func is not None:
                            func(f,**kwargs)
            else:
                if i == 0:
                    mvf = MVForecaster(*fs,**self.kwargs)
                    i += 1
                func_list(mvf,**kwargs)
        return tuple(fs) if i == 2 else mvf


from scalecast.AnomalyDetector import AnomalyDetector
from test_Forecaster import build_Forecaster
import matplotlib.pyplot as plt

def main():
    f = build_Forecaster(test_length = 0)
    detector = AnomalyDetector(f)
    detector.MonteCarloDetect('2010-01-01','2021-06-01')
    detector.plot_mc_results()
    plt.savefig('../../mc_anom.png')
    plt.close()
    detector.MonteCarloDetect_sliding(60,30)
    detector.labeled_anom
    detector.plot_anom(label=True)
    plt.savefig('../../mc_sliding_anom.png')
    plt.close()
    f = detector.adjust_anom(f=f,method='q')
    detector.EstimatorDetect(
        estimator='lstm',
        cilevel=.99,
        test_only=False,
        lags=24,
        epochs=25,
        validation_split=.2,
        shuffle=True,
        lstm_layer_sizes=(16,16,16),
        dropout=(0,0,0),
    )
    detector.plot_anom(label=True)
    plt.savefig('../../lstm_anom.png')
    plt.close()
    f = detector.WriteAnomtoXvars(f=f,drop_first=True)

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import shutil
import os
import sys

long_description = open('README.md', 'r', encoding="UTF-8").read()

setup(
  name = 'SCALECAST',
  version = '0.17.14',
  license='MIT',
  description="The practitioner's time series forecasting library",
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Michael Keith',
  author_email = 'mikekeith52@gmail.com',
  packages=find_packages('src'),
  package_dir={'': 'src'},
  package_data = {
    'scalecast':['grids/*'],
  },
  project_urls = {
    'GitHub': 'https://github.com/mikekeith52/scalecast',
    'Read the Docs': 'https://scalecast.readthedocs.io/en/latest/',
    'Examples': 'https://scalecast-examples.readthedocs.io/en/latest/',
  },
  url = 'https://github.com/mikekeith52/scalecast',
  keywords = ['FORECAST', 'SCALE', 'DYNAMIC', 'MACHINE LEARNING', 'APPLIED'],
  install_requires = [
    'scikit-learn',
    'statsmodels',
    'scipy',
    'eli5',
    'numpy',
    'pandas',
    'scipy',
    'matplotlib',
    'seaborn',
    'xgboost',
    'lightgbm',
    'catboost',
    'openpyxl',
  ],
)



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
import warnings
from sklearn.metrics import (
    r2_score,
    mean_squared_error,
    mean_absolute_percentage_error,
    mean_absolute_error,
)
from ._utils import _developer_utils

class metrics:
    @staticmethod
    def bias(a,f):
        """ Returns the total bias over a given forecast horizon. 
        When this is larger than 0, means aggregated predicted points are higher than actuals.
        Divide by the length of the forecast horizon to get average bias.

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.

        Returns:
            (float): The derived bias.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> metrics.bias(a,f) # returns 1
        """
        return np.sum(np.array(f) - np.array(a))

    @staticmethod
    def mape(a,f):
        """ Mean absolute percentage error (MAPE).

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.

        Returns:
            (float): The derived MAPE.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> metrics.mape(a,f)
        """
        return (
            np.nan if np.abs(a).min() == 0 
            else mean_absolute_percentage_error(a, f)
        )


    @staticmethod
    def r2(a,f):
        """ R-squared (R2).

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.

        Returns:
            (float): The derived R2.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> metrics.r2(a,f)
        """
        return r2_score(a, f)

    @staticmethod
    def mse(a,f):
        """ Mean squared error (MSE).

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.

        Returns:
            (float): The derived MSE.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> metrics.mse(a,f)
        """
        return mean_squared_error(a, f)

    @staticmethod
    def rmse(a,f):
        """ Root mean squared error (RMSE).

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.

        Returns:
            (float): The derived RMSE.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> metrics.rmse(a,f)
        """
        return mean_squared_error(a, f) ** 0.5

    @staticmethod
    def mae(a,f):
        """ Mean absolute error (MAE).

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.

        Returns:
            (float): The derived MAE.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> metrics.mae(a,f)
        """
        return mean_absolute_error(a, f)

    @staticmethod
    def smape(a,f):
        """ Symmetric mean absolute percentage error (sMAPE).
        Uses the same definition as used in the M4 competition.
        Does not multiply by 100.
        See https://ideas.repec.org/a/eee/intfor/v36y2020i1p54-74.html.

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.

        Returns:
            (float): The derived sMAPE.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> metrics.smape(a,f)
        """
        a = np.array(a)
        f = np.array(f)
        return (
            1/len(a) *
            np.sum(
                2*np.abs(f-a) / (
                    np.abs(a) + np.abs(f)
                )
            )
        )

    @staticmethod
    def mase(a,f,obs,m):
        """ Mean absolute scaled error (MASE).
        Uses the same definition as used in the M4 competition.
        See https://ideas.repec.org/a/eee/intfor/v36y2020i1p54-74.html.

        Args:
            a (list-like): The actuals over the forecast horizon.
            f (list-like): The predictions over the forecast horizon.
            obs (list-like): The actual observations used to create the forecast.
            m (int): The seasonal period.

        Returns:
            (float): The derived MASE.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> obs = [-5,-4,-3,-2,-1,0]
        >>> metrics.mase(a,f,obs,1)
        """
        a = np.array(a)
        f = np.array(f)
        avger = 1/len(a)
        num = np.sum(np.abs(f-a))
        davger = 1 / (len(obs) - m)
        denom = np.sum(
            np.abs(
                pd.Series(obs).diff(m).values[m:]
            )
        )
        return avger * (num / (davger * denom))

    @staticmethod
    def msis(a,uf,lf,obs,m,alpha=0.05):
        """ Mean scaled interval score (MSIS) for evaluating confidence intervals.
        Uses the same definition as used in the M4 competition.
        Lower values are better.
        See https://ideas.repec.org/a/eee/intfor/v36y2020i1p54-74.html.

        Args:
            a (list-like): The actuals over the forecast horizon.
            uf (list-like): The upper-forecast bound according to the confidence interval.
            lf (list-like): The lower-forecast bound according to the confidence interval.
            obs (list-like): The actual observations used to create the forecast.
            m (int): The seasonal period.
            alpha (float): Default 0.05. 0.05 for 95% confidence intervals, etc.

        Returns:
            (float): The derived MSIS.

        >>> from scalecast.util import metrics
        >>> a = [1,2,3,4,5]
        >>> f = [1,2,3,4,6]
        >>> uf = [1.5,2.5,3.5,4.5,6.5]
        >>> lf = [.5,1.5,2.5,3.5,5.5]
        >>> obs = [-5,-4,-3,-2,-1,0]
        >>> metrics.msis(a,uf,lf,obs,1) # returns a value of 5.0
        """
        a = np.array(a)
        uf = np.array(uf)
        lf = np.array(lf)
        avger = 1/len(a)
        num1 = uf-lf
        num2 = np.array([(
            (2/alpha*(lfs-acs)) if lfs > acs else
            (2/alpha*(acs-ufs)) if acs > ufs else
            0
        ) for acs, ufs, lfs in zip(a,uf,lf)])
        davger = 1/(len(obs) - m)
        denom = np.sum(np.abs(pd.Series(obs).diff(m).values[m:]))
        return avger * (np.sum(num1 + num2) / (davger * denom))

def plot_reduction_errors(f, ax = None, figsize=(12,6)):
    """ Plots the resulting error/accuracy of a Forecaster object where `reduce_Xvars()` method has been called
    with method = 'pfi' or method = 'shap'.
    
    Args:
        f (Forecaster): An object that has called the `reduce_Xvars()` method with method = 'pfi'.
        ax (Axis): Optional. The existing axis to write the resulting figure to.
        figsize (tuple): Default (12,6). The size of the resulting figure. Ignored when ax is not None.
        
    Returns:
        (Axis) The figure's axis.

    >>> from scalecast.Forecaster import Forecaster
    >>> from scalecast.util import plot_reduction_errors
    >>> import matplotlib.pyplot as plt
    >>> import seaborn as sns
    >>> import pandas as pd
    >>> import pandas_datareader as pdr
    >>> 
    >>> df = pdr.get_data_fred('HOUSTNSA',start='1900-01-01',end='2021-06-01')
    >>> f = Forecaster(y=df['HOUSTNSA'],current_dates=df.index)
    >>> f.set_test_length(.2)
    >>> f.generate_future_dates(24)
    >>> f.add_ar_terms(24)
    >>> f.integrate(critical_pval=.01)
    >>> f.add_seasonal_regressors('month',raw=False,sincos=True,dummy=True)
    >>> f.add_seasonal_regressors('year')
    >>> f.add_time_trend()
    >>> f.reduce_Xvars(method='pfi')
    >>> plot_reduction_errors(f)
    >>> plt.show()
    """
    dropped = f.pfi_dropped_vars
    errors = f.pfi_error_values
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    sns.lineplot(
        x=np.arange(0, len(dropped) + 1, 1), y=errors,
    )
    plt.xlabel("dropped Xvars")
    plt.ylabel("error")
    return ax

def backtest_metrics(
    backtest_results: list,
    mets = ['rmse'],
    mase = False, 
    msis = False,
    msis_alpha = 0.05,
    m = 1, 
    names = None,
) -> pd.DataFrame:
    """ Ingests the results output from `Pipeline.backtest()` and converts results to metrics.
    
    Args:
        backtest_results (list): The output returned from `Pipeline.backtest()` or `MVPipeline.backtest()`.
        mets (list): Default ['rmse']. A list of metrics to calculate.
            If the element is str type, must be taked from the `util.metrics` class 
            where the only two accepted arguments are a and f.
            If the element in the list is callable, must be a function that only accepts two arguments and returns a float.
        mase (bool): Default False.
            Whether to also calculate mase. Must specify seasonality in m.
        msis (bool): Default False.
            Whether to also calculate msis. Must specify seasonality in m. This will fail if confidence intervals were not evaluated.
        msis_alpha (float): Default 0.05. The level that confidence intervals were evaluated at. Ignored if msis is False.
        m (int): Default 1. The number of steps that count one seasonal cycle. Ignored if both of msis and mase is False.
        names (list): Optional. The names to assign each passed series. Ignored if there is only one passed series.
    
    Returns:
        (DataFrame): The metrics dataframe that gives info about each backtested series, model, and selected metric.

    >>> f1, f2, f3 = pipeline.fit_predict(f1,f2,f3)
    >>> backtest_results = pipeline.backtest(f1,f2,f3,n_iter=2,jump_back=12)
    >>> backtest_mets = backtest_metrics(
    >>>     backtest_results,
    >>>     mets = ['rmse','smape','r2','mae'],
    >>>     names=['UTUR','UNRATE','SAHMREALTIME'],
    >>>     mase = True,
    >>>     msis = True,
    >>>     m = 12,
    >>> )
    """
    m_ = m
    labels = names if names is not None else [f'Series{i}' for i in range(len(backtest_results))]
    res_dict = {k:None for k in labels}
    mets_str = [m if isinstance(m,str) else m.__name__ for m in mets]
    for h, s in enumerate(backtest_results):
        for i, m in enumerate(s):
            if i == 0:
                a_df = s[m][[c for c in s[m] if c.endswith('Vals')]]
                results = pd.DataFrame(
                    columns = [f'Iter{i}' for i in range(a_df.shape[1])],
                    index = pd.MultiIndex.from_product(
                            [
                                [m for m in backtest_results[0] if m not in ('Actuals','Obs')],
                                mets_str + ([] if not mase else ['mase']) + ([] if not msis else ['msis']),
                            ],
                            names = ['Model','Metric']
                    ),
                )
                a_df.columns = results.columns.to_list()
            elif i == 1:
                ob_df = s[m]
            else:
                f_df = s[m][[c for c in s[m] if c.endswith('Fcst')]]
                f_df.columns = results.columns.to_list()
                for met in mets_str:
                    for c in results:
                        results.loc[(m,met),c] = (
                            getattr(metrics,met)(
                                a = a_df[c],
                                f = f_df[c],
                            ) if isinstance(met,str) 
                            else met(
                                a_df[c],
                                f_df[c],
                            )
                        )
                if mase:
                    for c in results:
                        results.loc[(m,'mase'),c] = getattr(metrics,'mase')(
                            a = a_df[c],
                            f = f_df[c],
                            obs = ob_df[c].dropna(),
                            m = m_,
                        )
                if msis:
                    up_df = s[m][[c for c in s[m] if c.endswith('Upper')]]
                    lo_df = s[m][[c for c in s[m] if c.endswith('Lower')]]
                    up_df.columns = f_df.columns.to_list()
                    lo_df.columns = f_df.columns.to_list()
                    for c in results:
                        results.loc[(m,'msis'),c] = getattr(metrics,'msis')(
                            a = a_df[c],
                            uf = up_df[c],
                            lf = lo_df[c],
                            obs = ob_df[c].dropna(),
                            m = m_,
                            alpha = msis_alpha,
                        )
        res_dict[labels[h]] = results

    if len(res_dict) == 1:
        results_df = res_dict[labels[0]]
    else:
        res_list = []
        for k, df in res_dict.items():
            df['Series'] = k
            df = df.reset_index().set_index(['Series','Model','Metric'])
            res_list.append(df)
        results_df = pd.concat(res_list)

    results_df['Average'] = results_df.mean(axis=1)
    return results_df

def _backtest_plot(
    backtest_results,
    models = None,
    series = None,
    names = None,
    ci = False,
    ax=None,
    figsize=(12,6),
):
    """ (Coming soon). Plots the results from a backtested pipeline. If all default arguments are passed, the resulting plot can
    be messy, so it's recommended to limit some of the output by specifying arguments.

    Args:
        backtest_results (list): The output returned from `Pipeline.backtest()` or `MVPipeline.backtest()`.
        models (list): Which models to plot results for. By default, all models will be plotted.
        series (list): Which series to plot results for. This argument expects a list with an index position.
            Ex: [0,2,3] will plot series in the first, third, and fourth index positions of backtest_results.
        names (list): Optional. The names to give to the corresponding series. Must be the same length as series.
            If not specified, the series will be named 'Series1' - 'SeriesN'.
        ci (bool): Default False. Whether to plot confidence intervals. This will raise a warning if confidence intervals are not found.
        ax (Axis): Optional. A pre-existing axis to write the figure to.
        figsize (tuple): Default (12,6). The size of the resulting figure. Ignored when ax is specified.
    
    Returns:
        (Axis): The figure's axis.

    >>>
    >>>
    """
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

def break_mv_forecaster(mvf):
    """ Breaks apart an MVForecaster object and returns as many Foreaster objects as series loaded into the object.

    Args:
        mvf (MVForecaster): The object to break apart.

    Returns:
        (tuple[Forecaster]): A sequence of at least two Forecaster objects
    
    >>> from scalecast.MVForecaster import MVForecaster
    >>> from scalecast.util import break_mv_forecaster
    >>> 
    >>> f1, f2 = break_mv_forecaster(mvf)
    """
    from .Forecaster import Forecaster
    def convert_mv_hist(f, mvhist: dict, series_num: int):
        hist = {}
        for k, v in mvhist.items():
            hist[k] = {}
            for k2, v2 in v.items():
                if k2 in (""):
                    continue
                elif not isinstance(v2, dict) or k2 == "HyperParams":
                    hist[k][k2] = v2
                elif isinstance(v2, dict):
                    try:
                        hist[k][k2] = list(v2.values())[series_num]
                    except IndexError:
                        hist[k][k2] = []
            hist[k]["TestOnly"] = False
            hist[k]["LevelY"] = f.levely
        return hist

    mvf1 = mvf.deepcopy()

    set_len = (
        len(mvf1.series1['y']) 
        if not mvf1.current_xreg 
        else len(list(mvf1.current_xreg.values())[0])
    )
    to_return = []
    for s in range(mvf1.n_series):
        f = Forecaster(
            y=getattr(mvf1, f"series{s+1}")["y"].values[-set_len:],
            current_dates=mvf1.current_dates.values[-set_len:],
            integration=getattr(mvf1, f"series{s+1}")["integration"],
            levely=getattr(mvf1, f"series{s+1}")["levely"][-set_len:],
            init_dates=getattr(mvf1, f"series{s+1}")["init_dates"][-set_len:],
            future_dates=len(mvf1.future_dates),
            current_xreg={k:v.copy() for k,v in mvf1.current_xreg.items()},
            future_xreg={k:v.copy() for k,v in mvf1.future_xreg.items()},
            test_length=mvf1.test_length,
            validation_length=mvf1.validation_length,
            cis = mvf1.cis,
            cilevel = mvf1.cilevel,
        )
        f.history = convert_mv_hist(f, mvf1.history, s)
        to_return.append(f)

    return tuple(to_return)

def find_optimal_lag_order(mvf,train_only=False,**kwargs):
    """ Returns the otpimal lag order for a mutlivariate process using the statsmodels function:
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.vector_ar.var_model.VAR.select_order.html.
    The exogenous regressors are set based on Xvars loaded in the `MVForecaster` object.

    Args:
        mvf (MVForecaster): The MVForecaster object with series loaded to find the optimal order for
        train_only (bool): Default False. Whether to use the training data only in the test.
        **kwargs: Passed to the referenced statsmodels function

    Returns:
        (LagOrderResults): Lag selections.
    
    >>> from scalecast.Forecaster import Forecaster
    >>> from scalecast.MVForecaster import MVForecaster
    >>> from scalecast.util import find_optimal_lag_order
    >>> import pandas_datareader as pdr
    >>>
    >>> s1 = pdr.get_data_fred('UTUR',start='2000-01-01',end='2022-01-01')
    >>> s2 = pdr.get_data_fred('UNRATE',start='2000-01-01',end='2022-01-01')
    >>>
    >>> f1 = Forecaster(y=s1['UTUR'],current_dates=s1.index)
    >>> f2 = Forecaster(y=s2['UNRATE'],current_dates=s2.index)
    >>>
    >>> mvf = MVForecaster(f1,f2,names=['UTUR','UNRATE'])
    >>> lag_order_res = find_optimal_lag_order(mvf,train_only=True)
    >>> lag_order_aic = lag_order_res.aic # picks the best lag order according to aic
    """
    from statsmodels.tsa.vector_ar.var_model import VAR
    data = np.array(
        [getattr(mvf,f'series{i+1}')['y'].astype(float).values for i in range(mvf.n_series)],
    ).T

    if mvf.current_xreg:
        exog = pd.DataFrame(mvf.current_xreg).values
    else:
        exog = None

    if train_only:
        data = data[:-mvf.test_length]
        if exog is not None:
            exog = exog[:-mvf.test_length]

    model = VAR(data,exog=exog)

    return model.select_order(
        **kwargs,
    )

def find_optimal_coint_rank(mvf,det_order,k_ar_diff,train_only=False,**kwargs):
    """ Returns the optimal cointigration rank for a multivariate process using the function from statsmodels: 
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.vector_ar.vecm.select_coint_rank.html

    Args:
       mvf (MVForecaster): The MVForecaster object with series loaded to find the optimal rank for.
       train_only (bool): Default False. Whether to use the training data only in the test.
        **kwargs: Passed to the referenced statsmodels function.

    Returns:
        (CointRankResults): Object containing the cointegration rank suggested by the test and allowing a summary to be printed.

    >>> from scalecast.Forecaster import Forecaster
    >>> from scalecast.MVForecaster import MVForecaster
    >>> from scalecast.util import find_optimal_coint_rank
    >>> import pandas_datareader as pdr
    >>>
    >>> s1 = pdr.get_data_fred('UTUR',start='2000-01-01',end='2022-01-01')
    >>> s2 = pdr.get_data_fred('UNRATE',start='2000-01-01',end='2022-01-01')
    >>>
    >>> f1 = Forecaster(y=s1['UTUR'],current_dates=s1.index)
    >>> f2 = Forecaster(y=s2['UNRATE'],current_dates=s2.index)
    >>>
    >>> mvf = MVForecaster(f1,f2,names=['UTUR','UNRATE'])
    >>> coint_res = find_optimal_coint_rank(mvf,det_order=-1,k_ar_diff=8,train_only=True)
    >>> print(coint_res) # prints a report
    >>> rank = coint_res.rank # best rank
    """
    from statsmodels.tsa.vector_ar.vecm import select_coint_rank
    data = np.array(
        [getattr(mvf,f'series{i+1}')['y'].values for i in range(mvf.n_series)],
    ).T
    if train_only:
        data = data[:-mvf.test_length]
    
    return select_coint_rank(
        data,
        det_order=det_order,
        k_ar_diff=k_ar_diff,
        **kwargs,
    )

def find_statistical_transformation(
    f,
    goal=['stationary'],
    train_only=False,
    critical_pval=.05,
    log = True,
    m='auto',
    adf_kwargs = {},
    **kwargs,
):
    """ Finds a set of transformations to achieve stationarity or seasonal adjustment, based on results from statistical tests.

    Args:
        f (Forecaster): The object that stores the series to test.
        goal (list-like): Default ['stationary']. One or multiple of 'stationary', 'seasonally_adj'. 
            Other options may be coming in the future.
            If more than one goal is passed, will try to satisfy all goals in the order passed.
            For stationary: uses an Augmented Dickey-Fuller test to determine if the series is stationary.
            If not stationary, returns a diff transformation and log transformation if log is True.
            For seasonall_adj: uses seasonal auto_arima to find the optimal seasonal diff.
        train_only (bool): Default False. Whether to use train set only in all statistical tests.
        log (bool): Default True. Whether to log and difference the series if it is found to be non-stationary or just difference.
            This will set itself to False if the lowest observed series value is 0 or lower.
        critical_pval (float): Default 0.05. The cutoff p-value to use to determine statistical signficance in the 
            Augmented Dickey-Fuller test and to run the auto_arima selection (substitutes for `alpha` arg).
        m (str or int): Default 'auto': The number of observations that counts one seasonal step.
            When 'auto', uses the M4 competition values: 
            for Hourly: 24, Monthly: 12, Quarterly: 4. everything else gets 1 (no seasonality assumed)
            so pass your own values for other frequencies.
        adf_kwargs (dict): Default {}. Keyword args to pass to the Augmented Dickey-Fuller test function.
            See the `maxlag`, `regression`, and `autolag` arguments from
            https://www.statsmodels.org/dev/generated/statsmodels.tsa.stattools.adfuller.html.
        **kwargs: Passed to the auto_arima() function when searching for optimal seasonal diff.

    Returns:
        (Transformer, Reverter): A `Transformer` object with the identified transforming functions and
        the `Reverter` object with the `Transformer` counterpart functions.

    >>> from scalecast.Forecaster import Forecaster
    >>> from scaleast.Pipeline import Pipeline, Transformer, Reverter
    >>> from scalecast.util import find_statistical_transformation
    >>> import pandas_datareader as pdr
    >>> 
    >>> def forecaster(f):
    >>>     f.add_covid19_regressor()
    >>>     f.auto_Xvar_select(cross_validate=True)
    >>>     f.set_estimator('mlr')
    >>>     f.manual_forecast()
    >>> df = pdr.get_data_fred(
    >>>     'HOUSTNSA',
    >>>     start='1959-01-01',
    >>>     end='2022-08-01'
    >>> )
    >>> f = Forecaster(
    >>>     y=df['HOUSTNSA'],
    >>>     current_dates=df.index,
    >>>     future_dates=24,
    >>>     test_length = .2,
    >>> )
    >>> f.set_validation_length(24)
    >>> transformer, reverter = find_statistical_transformation(
    >>>     f,
    >>>     goal=['stationary','seasonally_adj'],
    >>>     train_only=True,
    >>>     critical_pval = .01,
    >>> )
    >>> print(reverter) # see what transformers and reverters were chosen
    >>> pipeline = Pipeline(
    >>>   steps = [
    >>>       ('Transform',transformer),
    >>>       ('Forecast',forecaster),
    >>>       ('Revert',reverter),
    >>>   ],
    >>> )
    >>> f = pipeline.fit_predict(f)
    """
    from . import SeriesTransformer
    from . import Pipeline
    from .auxmodels import auto_arima
    def make_stationary(f,train_only,critical_pval,log,adf_kwargs,**kwargs):
        transformers = []
        stationary = f.adf_test(
            train_only=train_only,
            full_res=False,
            critical_pval=critical_pval,
            **adf_kwargs
        )
        if not stationary:
            if log and f.y.min() > 0:
                transformers += [('LogTransform',)]
            transformers += [('DiffTransform',1)]  
        return transformers

    def seasonally_adj(f,train_only,critical_pval,log,adf_kwargs,**kwargs):
        transformers = []
        if m == 1:
            return transformers

        auto_arima(f,m=m,seasonal=True,alpha=critical_pval,**kwargs)
        I = f.auto_arima_params['seasonal_order'][1]
        if I > 0:
            for i in range(1,I+1):
                transformers += [('DiffTransform',m)]

        return transformers

    f = f.deepcopy()
    transformer = SeriesTransformer.SeriesTransformer(f)

    m = _developer_utils._convert_m(m,f.freq)

    possible_args = {
        'stationary':make_stationary,
        'seasonally_adj':seasonally_adj,
    }
    bad_args = [g for g in goal if g not in possible_args]
    if len(bad_args) > 0:
        raise ValueError(f'Values passed to goal arg cannot be used: {bad_args}.')

    transformers = []
    reverters = []
    for g in goal:
        t = possible_args[g](
            f,
            train_only=train_only,
            critical_pval=critical_pval,
            log=log,
            adf_kwargs=adf_kwargs,
            **kwargs
        )
        transformers += t
    for t in transformers[::-1]:
        r = (t[0].replace('Transform','Revert'),)
        if len(t) > 1:
            r += t[1:]
        reverters.append(r)

    final_transformer = Pipeline.Transformer(transformers = transformers)
    final_reverter = Pipeline.Reverter(
        reverters = reverters, 
        base_transformer = final_transformer,
    )
    return final_transformer, final_reverter

@_developer_utils.log_warnings
def find_optimal_transformation(
    f,
    estimator='mlr',
    monitor=None,
    lags='auto',
    try_order = ['detrend','seasonal_adj','boxcox','first_diff','first_seasonal_diff','scale'],
    boxcox_lambdas = [-0.5,0,0.5],
    detrend_kwargs = [{'poly_order':1},{'poly_order':2}],
    scale_type = ['Scale','MinMax'],
    scale_on_train_only = False,
    m='auto',
    model = 'add',
    **kwargs,
):
    """ Finds a set of transformations based on what maximizes forecast accuracy on some out-of-sample metric.
    Works by comparing each transformation individually and stacking the set of transformations that leads to the best
    performance. The estimator only uses series lags as inputs. When an attempted transformation fails, a warning is logged.

    Args:
        f (Forecaster): The Forecaster object that contains the series that will be transformed.
        estimator (str): One of _estimators_. Default 'mlr'. The estimator to use to choose the best 
            transformations with.
        monitor (str): One of Forecaster.determine_best_by except 'ValidationMetricValue'. 
            Default is 'TestSet' + first metric in Forecaster.metrics 
            ('TestSetRMSE' is the default if no metric info has been manually specified).
            Because 'ValidationMetricValue' changes in scale based on the transformation taken, 
            the metrics to monitor are limited to in-sample and test-set metrics.
            'TestSetRMSE' and 'LevelTestSetRMSE' are the same in this function, same with all level and non-level counterparts.
        lags (str or int): Default 'auto'. The number of lags that will be used as inputs for the estimator.
            If 'auto', uses the value passed or assigned to m (one seasonal cycle). 
            If multiple values passed to m, uses the first.
        try_order (list-like): Default ['detrend','seasonal_adj','boxcox','first_diff','first_seasonal_diff','scale'].
            The transformations to try and also the order to try them in.
            Changing the order here can change the final transformations derived, since level will 
            be compared to the first transformation and if it is found to be better than level, it will
            carry over to be tried in conjunction with the next transformation and so forth.
            The default list contains all possible transformations for this function.
        boxcox_lambdas (list-like): Default [-0.5,0,0.5].
            The lambda values to try for a boxcox transformation.
            0 means natural log. Only up to one boxcox transformation will be selected.
        detrend_kwargs (list-like[dict]): Default [{'poly_order':1},{'poly_order':2}].
            The types of detrending to try. Only up to one one detrender will be selected.
        scale_type (list-like): Default ['Scale','MinMax']. The type of scaling to try.
            Only up to one scaler will be selected.
            Must exist a `SeriesTranformer.{scale_type}Transform()` function for this to work.
        scale_on_train_only (bool): Default False. Whether to fit the scaler on the training set only.
        m (str, int, list[int]): Default 'auto'. The number of observations that counts one seasonal step.
            When 'auto', uses the M4 competition values: 
            for Hourly: 24, Monthly: 12, Quarterly: 4. everything else gets 1 (no seasonality assumed)
            so pass your own values for other frequencies. If m == 1, no first seasonal difference will be tried.
            If list, multiple seasonal differences can be tried and up to that many seasonal differences can be selected.
        model (str): Default 'add'. One of {"additive", "add", "multiplicative", "mul"}.
            The type of seasonal component. Only relevant for the 'seasonal_adj' option in try_order.
        **kwargs: Passed to the `Forecaster.manual_forecast()` function and possible values change based on which
            estimator is used.

    Returns:
        (Transformer, Reverter): A `Transformer` object with the identified transforming functions and
        the `Reverter` object with the `Transformer` counterpart functions.

    >>> from scalecast.Forecaster import Forecaster
    >>> from scaleast.Pipeline import Pipeline, Transformer, Reverter
    >>> from scalecast.util import find_optimal_transformation
    >>> import pandas_datareader as pdr
    >>> 
    >>> def forecaster(f):
    >>>     f.add_covid19_regressor()
    >>>     f.auto_Xvar_select(cross_validate=True)
    >>>     f.set_estimator('mlr')
    >>>     f.manual_forecast()
    >>> df = pdr.get_data_fred(
    >>>     'HOUSTNSA',
    >>>     start='1959-01-01',
    >>>     end='2022-08-01'
    >>> )
    >>> f = Forecaster(
    >>>     y=df['HOUSTNSA'],
    >>>     current_dates=df.index,
    >>>     future_dates=24,
    >>>     test_length = .2, # this will be monitored for performance
    >>> )
    >>> f.set_validation_length(24)
    >>> transformer, reverter = find_optimal_transformation(f)
    >>> print(reverter) # see what transformers and reverters were chosen
    >>> pipeline = Pipeline(
    >>>   steps = [
    >>>       ('Transform',transformer),
    >>>       ('Forecast',forecaster),
    >>>       ('Revert',reverter),
    >>>   ],
    >>> )
    >>> f = pipeline.fit_predict(f)
    """
    from . import Pipeline
    from ._Forecaster_parent import ForecastError
    def forecaster(f):
        f.add_ar_terms(lags)
        f.set_estimator(estimator)
        f.manual_forecast(**kwargs)

    def make_pipeline_fit_predict(f,transformer,reverter):
        tr = Pipeline.Transformer(transformers=transformer)
        re = Pipeline.Reverter(reverters=reverter,base_transformer=tr)
        pipeline = Pipeline.Pipeline(
            steps = [
                ('Transform',tr),
                ('Forecast',forecaster),
                ('Revert',re)
            ],
        )
        return pipeline.fit_predict(f)

    if f.test_length == 0:
        raise ForecastError(
            'find_optimal_transformation() only works if a test length' 
            ' above 0 is specified in the Forecaster object.'
            ' Try calling the set_test_length(...) method.'
        )

    f = f.deepcopy()
    f.drop_all_Xvars()
    f.history = {}

    m = _developer_utils._convert_m(m,f.freq)
    lags = m if lags == 'auto' and not hasattr(m,'__len__') else m[1] if lags == 'auto' else lags
    forecaster(f)
    
    monitor = 'TestSet' + list(f.metrics.keys())[0].upper() if monitor is None else monitor
    level_met = f.export('model_summaries')[monitor].values[0]
    level_met = -level_met if monitor.endswith('R2') else level_met

    final_transformer = []
    final_reverter = []

    exception_types = (IndexError,AttributeError,ValueError,ZeroDivisionError,ForecastError) # errors to pass over

    for tr in try_order:
        if tr == 'boxcox':
            def boxcox_tr(x,lmbda):
                return [(i**lmbda - 1) / lmbda for i in x] if lmbda != 0 else [np.log(i) for i in x]
            def boxcox_re(x,lmbda):
                return [(i*lmbda + 1)**(1/lmbda) for i in x] if lmbda != 0 else [np.exp(i) for i in x]
            for i, lmbda in enumerate(boxcox_lambdas):
                transformer = final_transformer[:]
                reverter = final_reverter[:]
                if i == 0:
                    met = level_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                try:
                    transformer.append(('Transform',boxcox_tr,{'lmbda':lmbda}))
                    reverter.reverse(); reverter.append(('Revert',boxcox_re,{'lmbda':lmbda})); reverter.reverse()
                    f = make_pipeline_fit_predict(f,transformer,reverter)
                    comp_met = f.export('model_summaries')[monitor].values[0]
                    comp_met = -comp_met if monitor.endswith('R2') else comp_met
                    if comp_met < met:
                        met = comp_met
                        best_transformer = transformer[:]
                        best_reverter = reverter[:]
                except exception_types as e:
                    warnings.warn(f'Lambda value of {lmbda} cannot be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        elif tr == 'detrend':
            for i, kw in enumerate(detrend_kwargs):
                transformer = final_transformer[:]
                reverter = final_reverter[:]
                if i == 0:
                    met = level_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                try:
                    transformer.append(('DetrendTransform',kw))
                    reverter.reverse(); reverter.append(('DetrendRevert',)); reverter.reverse()
                    f = make_pipeline_fit_predict(f,transformer,reverter)
                    comp_met = f.export('model_summaries')[monitor].values[0]
                    comp_met = -comp_met if monitor.endswith('R2') else comp_met
                    if comp_met < met:
                        met = comp_met
                        best_transformer = transformer[:]
                        best_reverter = reverter[:]
                except exception_types as e:
                    warnings.warn(f'Detrend_kwargs {kw} cannot be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        elif tr == 'first_diff':
            met = level_met
            transformer = final_transformer[:]
            reverter = final_reverter[:]
            best_transformer = transformer[:]
            best_reverter = reverter[:]
            try:
                transformer.append(('DiffTransform',1))
                reverter.reverse(); reverter.append(('DiffRevert',1)); reverter.reverse()
                f = make_pipeline_fit_predict(f,transformer,reverter)
                comp_met = f.export('model_summaries')[monitor].values[0]
                comp_met = -comp_met if monitor.endswith('R2') else comp_met
                if comp_met < met:
                    met = comp_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
            except exception_types as e:
                warnings.warn(f'Series first difference could not be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        elif tr in ('seasonal_adj', 'first_seasonal_diff'):
            if not hasattr(m,'__len__'):
                m = [m]
            for mi in m:
                if mi > 1:
                    met = level_met
                    transformer = final_transformer[:]
                    reverter = final_reverter[:]
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                    try:
                        transformer += (
                            [('DiffTransform',mi)] if tr == 'first_seasonal_diff'
                            else [('DeseasonTransform',{'m':mi,'model':model})]
                        )
                        reverter.reverse()
                        reverter += (
                            [('DiffRevert',mi)] if tr == 'first_seasonal_diff' 
                            else [('DeseasonRevert',)]
                        )
                        reverter.reverse()
                        f = make_pipeline_fit_predict(f,transformer,reverter)
                        comp_met = f.export('model_summaries')[monitor].values[0]
                        comp_met = -comp_met if monitor.endswith('R2') else comp_met
                        if comp_met < met:
                            met = comp_met
                            best_transformer = transformer[:]
                            best_reverter = reverter[:]
                    except exception_types as e:
                        warnings.warn(f'Series seasonal adjustment could not be evaluated. error: {e}')
                    final_transformer = best_transformer[:]
                    final_reverter = best_reverter[:]
                    level_met = met
                else:
                    warnings.warn('Seasonal differences and adjustments cannot be evaluated when m = 1.')
        elif tr == 'scale':
            for i, s in enumerate(scale_type):
                transformer = final_transformer[:]
                reverter = final_reverter[:]
                if i == 0:
                    met = level_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                try:
                    transformer.append((f'{s}Transform',{'train_only':scale_on_train_only}))
                    reverter.reverse(); reverter.append((f'{s}Revert',)); reverter.reverse()
                    f = make_pipeline_fit_predict(f,transformer,reverter)
                    comp_met = f.export('model_summaries')[monitor].values[0]
                    comp_met = -comp_met if monitor.endswith('R2') else comp_met
                    if comp_met < met:
                        met = comp_met
                        best_transformer = transformer[:]
                        best_reverter = reverter[:]
                except exception_types as e:                 
                    warnings.warn(f'{s} scaler cannot be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        else:
            warnings.warn(f'Value: {tr} found in the try_order list cannot be used and will be skipped.')
    
    final_transformer = Pipeline.Transformer(transformers = final_transformer)
    final_reverter = Pipeline.Reverter(reverters = final_reverter,base_transformer = final_transformer)
    return final_transformer, final_reverter 



from functools import wraps
import logging
import warnings
import numpy as np
from scipy import stats

class _developer_utils:
    @staticmethod
    def log_warnings(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings(record=True) as warn_list:
                warnings.simplefilter("always")
                result = func(*args, **kwargs)
                for warn in warn_list:
                    logging.warning(warn.message)
            return result
        return wrapper
    
    @staticmethod
    def descriptive_assert(statement, ErrorType, error_message):
        # descriptive assert statement for descriptive exception raising
        try:
            assert statement
        except AssertionError:
            raise ErrorType(error_message)

    @staticmethod      
    def _return_na_if_len_zero(y,pred,func):
        return np.nan if len(pred) == 0 else func(y,pred)

    @staticmethod
    def _set_ci_step(f,s):
        return stats.norm.ppf(1 - (1 - f.cilevel) / 2) * s

    @staticmethod
    def _check_train_only_arg(f, train_only):
        _developer_utils.descriptive_assert(
            isinstance(train_only, bool), ValueError, f"train_only must be True or False, got {train_only} of type {type(train_only)}."
        )
        _developer_utils.descriptive_assert(
            not train_only or f.test_length > 0, ValueError, "train_only cannot be True when test_length is 0."
        )

    @staticmethod
    def _check_if_correct_estimator(estimator,possible_estimators):
        _developer_utils.descriptive_assert(
            estimator in possible_estimators,
            ValueError,
            f"estimator must be one of {possible_estimators}, got {estimator}.",
        )

    @staticmethod
    def _warn_about_not_finding_cis(m):
        warnings.warn(
            f'Confidence intervals not found for {m}. '
            'To turn on confidence intervals for future evaluated models, call the eval_cis() method.'
            ,category=Warning
        )

    @staticmethod
    def _convert_m(m,freq):
        if m == 'auto':
            if freq is not None:
                if freq.startswith('M'):
                    return 12
                elif freq.startswith('Q'):
                    return 4
                elif freq.startswith('H'):
                    return 24
                else:
                    return 1
            else:
                return 1
        return m

    @staticmethod
    def _determine_best_by(metrics):
        return [
            'TestSet' + m.upper() for m in metrics
        ] + [
            'InSample' + m.upper() for m in metrics
        ] + [
            'LevelTestSet' + m.upper() for m in metrics
        ] + [
            'LevelInSample' + m.upper() for m in metrics
        ] + ['ValidationMetricValue']


__version__ = '0.17.14'

from .util import metrics
import inspect
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import GradientBoostingRegressor, RandomForestRegressor
from xgboost import XGBRegressor
from lightgbm import LGBMRegressor
from catboost import CatBoostRegressor
from sklearn.svm import SVR
from sklearn.neighbors import KNeighborsRegressor
from sklearn.linear_model import (
    LinearRegression,
    ElasticNet,
    Lasso,
    Ridge,
    SGDRegressor,
)
from sklearn.preprocessing import (
    MinMaxScaler, 
    StandardScaler, 
    Normalizer,
)

def _none(x):
    return x

__sklearn_imports__ = {
    "catboost": CatBoostRegressor,
    "elasticnet": ElasticNet,
    "gbt": GradientBoostingRegressor,
    "knn": KNeighborsRegressor,
    "lasso": Lasso,
    "lightgbm": LGBMRegressor,
    "mlp": MLPRegressor,
    "mlr":LinearRegression,
    "rf": RandomForestRegressor,
    "ridge": Ridge,
    "sgd": SGDRegressor,
    "svr": SVR,
    "xgboost": XGBRegressor,
}
# estimators
__sklearn_estimators__ = list(__sklearn_imports__.keys())
# to add non-sklearn models, add to the list below
# sklearn estimators go to the _sklearn_imports_ dict at the top
__non_sklearn_estimators__ = [
    "arima",
    "hwes",
    "prophet",
    "silverkite",
    "rnn",
    "lstm",
    'naive',
    "theta",
    "combo",
]
__estimators__ = __sklearn_estimators__ + __non_sklearn_estimators__
__cannot_be_tuned__ = ["combo", "rnn", "lstm", "naive"]
__can_be_tuned__ = [m for m in __estimators__ if m not in __cannot_be_tuned__]

# only methods with two arguments (a, f) can be used in validation
__metrics__ = {
    name:method for name, method in inspect.getmembers(metrics, inspect.isroutine) 
    if not name.startswith('_') and len(inspect.signature(method).parameters) == 2
}
__normalizer__ = {
    "minmax":MinMaxScaler, 
    "normalize":Normalizer, 
    "scale":StandardScaler, 
    None:_none,
}
# i do it this way to make mvforecaster work a little better
__colors__ = [
    "#FFA500",
    "#DC143C",
    "#00FF7F",
    "#808000",
    "#BC8F8F",
    "#A9A9A9",
    "#8B008B",
    "#FF1493",
    "#FFDAB9",
    "#20B2AA",
    "#7FFFD4",
    "#A52A2A",
    "#DCDCDC",
    "#E6E6FA",
    "#BDB76B",
    "#DEB887",
] * 10
__series_colors__ = [
    "#0000FF",
    "#00FFFF",
    "#7393B3",
    "#088F8F",
    "#0096FF",
    "#F0FFFF",
    "#00FFFF",
    "#5D3FD3",
    "#191970",
    "#9FE2BF",
] * 10
# keywords that are passed to _bank_history() that I don't want to be recognized as hyperparams
__not_hyperparams__ = ["Xvars", "normalizer", "tuned", "plot_loss", "plot_loss_test"]